#ifndef __BIT_VEC
#define __BIT_VEC

#include <iostream>

class BitVec
{
friend std::ostream& operator<<(std::ostream&, const BitVec&);

public:
  BitVec(const BitVec&);
  BitVec(uint64_t _num_trans, uint64_t _min_item, uint64_t _max_item);
  ~BitVec();

  void set_one(int transaction, int item);

  void set_zero(int transaction, int item);
  
  bool get(int transaction, int item) const;

  uint64_t num_bits_set() const;

  uint64_t num_trans() const;
  uint64_t min_item() const;
  uint64_t max_item() const;

  void transpose();

  uint64_t get_bit_loc(int transaction, int item) const;

  void write_to_trans_file(const std::string&) const;

protected:

  void *db;
  uint64_t _num_trans;
  uint64_t _min_item;
  uint64_t _max_item;
  uint64_t _item_width;
};

inline
uint64_t BitVec::get_bit_loc(int transaction, int item) const
{
  return ((uint64_t)(transaction - 1) * _item_width) + (item - _min_item);
}

inline
void BitVec::set_one(int transaction, int item)
{
  //const uint64_t bit_loc = (uint64_t)(transaction - 1) * _item_width + (item - min_item);
  //static_cast<uint64_t*>(db)[bit_loc >> 6] |= ((uint64_t)1 << (bit_loc & (uint64_t)63));
  const uint64_t bit_loc = get_bit_loc(transaction, item);
  static_cast<unsigned char*>(db)[bit_loc >> 3] |= (1 << (bit_loc & 7));
}

inline
void BitVec::set_zero(int transaction, int item)
{
  const uint64_t bit_loc = get_bit_loc(transaction, item);
  static_cast<unsigned char*>(db)[bit_loc >> 3] &= ~(1 << (bit_loc & 7));
}
  
inline
bool BitVec::get(int transaction, int item) const
{
  //const uint64_t bit_loc = (uint64_t)(transaction - 1) * _item_width + (item - min_item);
  //return (static_cast<uint64_t*>(db)[bit_loc >> 6] >> (bit_loc & (uint64_t)63)) & (uint64_t)1;
  const uint64_t bit_loc = get_bit_loc(transaction, item);
  //std::cout << "get(" << transaction << ", " << item << ") using bit_loc " << bit_loc << "\n";
  return (static_cast<unsigned char*>(db)[bit_loc >> 3] >> (bit_loc & 7)) & 1;
}


#endif
