#include <limits>
#include "Graph.h"

//#define DEBUG

Graph::Graph(int size) {
	V=size;

	//initializing order
	order=new int [V];
	for (int i=0; i<V; i++){
		order[i]=i;
	}

/*	//allocate memory for matrices and initialize them
	adjacency_matrix=new unsigned short int *[V];
	for (int i=0; i<V; i++){
		adjacency_matrix[i]=new unsigned short int[V];
		for (int j=0; j<V; j++){
			adjacency_matrix[i][j]=0;
		}
	}
*/
	adjacency_list=new edge [V];
}

Graph::Graph() {
	adjacency_list = NULL;
	order = NULL;

}

Graph::Graph(const Graph& g)
{
	V = g.V;
	E = g.E;
	adjacency_list = new edge[V];
        for (int i=0; i<V; ++i)
          adjacency_list[i] = g.adjacency_list[i];
	
	order = new int[V];
        for (int i=0; i<V; ++i)
          order[i] = g.order[i];
	
}

Graph& Graph::operator=(const Graph& g)
{
	V = g.V;
	E = g.E;
	adjacency_list = new edge[V];
        for (int i=0; i<V; ++i)
          adjacency_list[i] = g.adjacency_list[i];
	
	order = new int[V];
        for (int i=0; i<V; ++i)
          order[i] = g.order[i];
	return *this;
}

int Graph::sizeE() {
	E=0;

/*	
	for (int i=0; i<V-1; i++){
		for (int j=i+1; j<V; j++){
			if (adjacency_matrix[i][j]>0)
				E++;
		}
	}
*/
	for (int i=0; i<V; i++){
		E=E+adjacency_list[i].size();
	}


	return E;
}

Graph::~Graph(){

/*
	for (int i=0; i<V; i++){
		delete[] adjacency_matrix[i];
	}
	delete[] adjacency_matrix;
*/
    if (adjacency_list != NULL)
	delete[] adjacency_list;
    /*
    if (order != NULL)
        delete[] order;
    */
}

int Graph::Graphcost(){
//	int i=0;
	int cost=0;
	int* reverse_mapping;
	reverse_mapping=new int[V];
	for (int i=0; i<V; i++){
		reverse_mapping[order[i]]=i;
	}

/*	
	while (i<V-1){
		for (int j=i+1; j<V; j++){
			if (adjacency_matrix[i][j]>0)
				cost+=adjacency_matrix[i][j]*int(fabs(reverse_mapping[j]-reverse_mapping[i]));
		}
		i++;
	}
*/
	for (int i=0; i<V; i++)
		for (edge::iterator it=adjacency_list[i].begin(); it!=adjacency_list[i].end(); it++)
			cost+=(int)(*it).second*(int)(fabs((*it).first-i));
	
	delete[] reverse_mapping;

	return cost;
}

void Graph::Graph_display(){
	cout<<"Graph is displayed as follow:"<<endl;

/*
	for (int i=0; i<V; i++){
		for (int j=0; j<V; j++)
			if (adjacency_matrix[i][j]>0)
				cout<<"("<<i<<","<<j<<")|"<<adjacency_matrix[i][j]<<";";
		cout<<endl;
	}
*/

	for(int i=0; i<V; i++){
		for (edge::iterator it=adjacency_list[i].begin(); it!=adjacency_list[i].end(); it++){
			cout<<"("<<i<<","<<it->first<<")|"<<it->second<<";";
			}
	//	cout<<endl;
	}
	
}

Hypergraph::Hypergraph(int vsize, int esize) {
	V=vsize;
	E=esize;

	//initializing order
	order=new int [V];
	for (int i=0; i<V; i++){
		order[i]=i;
	}

	//allocate memory for matrices and initialize them
	transaction_matrix=new BitVec (E, 1, V);
//	edge_head_tail_matrix=new int *[E];
	for (int i=0; i<E; i++){
		for (int j=0; j<V; j++){
			transaction_matrix->set_zero(i+1,j+1);
		}
//		edge_head_tail_matrix[i]=new int[2];
//		edge_head_tail_matrix[i][0]=-1;
//		edge_head_tail_matrix[i][1]=-1;

	}
}

Hypergraph::~Hypergraph(){

	delete transaction_matrix;
//	delete[] edge_head_tail_matrix;
	delete[] order;
}

int Hypergraph::Hypercost(){
	int** edge_head_tail_matrix; //edges as row. Two columns: head and tail
	edge_head_tail_matrix=new int*[E];
	
	for (int i=0; i<E; i++){
		edge_head_tail_matrix[i]=new int[2];
		edge_head_tail_matrix[i][0]=-1;
		edge_head_tail_matrix[i][1]=-1;
	}

	//fill up edge_head_tail_matrix
	for (int i=0; i<V; i++){
		for (int j=0; j<E; j++){
			if (transaction_matrix->get(j+1,order[i]+1)==1) {
	//			cout<<"j:"<<j<<"order[i]:"<<order[i]<<endl;
				if (edge_head_tail_matrix[j][0]==-1){
	//				cout<<"assign head:"<<i<<endl;
					edge_head_tail_matrix[j][0]=i;
				}
				edge_head_tail_matrix[j][1]=i;
			}

		}
	}

	//calculate cost
	int cost=0;
	for (int i=0; i<E; i++){
		cost+=edge_head_tail_matrix[i][1]-edge_head_tail_matrix[i][0];
	}

	
	for (int i=0; i<E; i++){
		delete[] edge_head_tail_matrix[i];
	}
	delete[] edge_head_tail_matrix;

	return cost;
}

Graph Hypergraph::Hypertograph_Line(int* vertex_mapping_table, int* reverse_vertex_mapping_table, vector<int>& isolated_vertices){

	int* reverse_mapping;
	reverse_mapping=new int[V];
	for (int i=0; i<V; i++){
		reverse_mapping[order[i]]=i;
	}

	BitVec* sparse_matrix=new BitVec(V,1,V);
	
/*
	//initializing sparse_matrix//not necessary
	for (int i=0; i<V; i++)
		for (int j=0; j<V; j++)
			sparse_matrix->set_zero(i+1,j+1);
*/
	
	//calculate sparse matrix
	for (int j=0; j<E; j++){
		hyperedge heg;
		for (int i=0; i<V; i++)
			if (transaction_matrix->get(j+1,i+1)==1){
				heg.insert(pair<ranking, vertex>(reverse_mapping[i],i));
			}
		if (heg.empty())
			continue;
		map<ranking,vertex>::iterator it1,it2;
		it1=heg.begin();
		it2=heg.begin();
		it2++;
		#ifdef DEBUG		
		cout<<"forming line:";
		#endif
		sparse_matrix->set_one((*it1).second+1, (*it1).second+1);
		while (it2!=heg.end()){
//			g.adjacency_matrix[(*it1).second][(*it2).second]++;//symmetric
//			g.adjacency_matrix[(*it2).second][(*it1).second]++;
			#ifdef DEBUG
			cout<<"("<<(*it1).second<<","<<(*it2).second<<"),";
			#endif
			sparse_matrix->set_one((*it1).second+1, (*it2).second+1);
			sparse_matrix->set_one((*it2).second+1, (*it1).second+1);
			it1++;
			it2++;
		}
		//cout<<endl;
	}

	//eliminate isolated vertices

	#ifdef DEBUG
	cout<<"forming vertex mapping table:";
	#endif
	int vertex_mapping_pointer=0;
	for (int i=0; i<V; i++){
		for (int j=0; j<V; j++){
			if (sparse_matrix->get(i+1,j+1)>0 ){
				#ifdef DEBUG
			   	cout<<"("<<i<<","<<j<<"),";
				#endif
				vertex_mapping_table[vertex_mapping_pointer]=i;
				vertex_mapping_pointer++;
				break;
			}
			if (j==V-1)
				isolated_vertices.push_back(i);
		}
	}
	#ifdef DEBUG
	cout<<endl;
	#endif
	for (unsigned int i=0; i<isolated_vertices.size(); i++){
		vertex_mapping_table[vertex_mapping_pointer]=isolated_vertices[i];
		vertex_mapping_pointer++;
	}
	
	for (int i=0; i<V; i++)
		reverse_vertex_mapping_table[vertex_mapping_table[i]]=i;
	delete sparse_matrix;
	
	cout<<"V:"<<V<<endl;
	cout<<"compressed vertices:"<<isolated_vertices.size()<<endl;
	cout<<"delete sparse_matrix successfully"<<endl;
	
	//Create compressed graph
	Graph g(V-isolated_vertices.size());
	for (int j=0; j<E; j++){
		hyperedge heg;
		for (int i=0; i<V; i++)
			if (transaction_matrix->get(j+1,i+1)==1){
				heg.insert(pair<ranking, vertex>(reverse_mapping[i],i));
			}
		if (heg.empty())
			continue;
		map<ranking,vertex>::iterator it1,it2;
		it1=heg.begin();
		it2=heg.begin();
		it2++;
		while (it2!=heg.end()){
//			g.adjacency_matrix[reverse_vertex_mapping_table[(*it1).second]][reverse_vertex_mapping_table[(*it2).second]]++;//symmetric
//			g.adjacency_matrix[reverse_vertex_mapping_table[(*it2).second]][reverse_vertex_mapping_table[(*it1).second]]++;
			int large=0;
			int small=0;
			if (reverse_vertex_mapping_table[(*it1).second]>reverse_vertex_mapping_table[(*it2).second]){
				small=reverse_vertex_mapping_table[(*it2).second];
				large=reverse_vertex_mapping_table[(*it1).second];
			}else if (reverse_vertex_mapping_table[(*it1).second]<reverse_vertex_mapping_table[(*it2).second]){
				small=reverse_vertex_mapping_table[(*it1).second];
				large=reverse_vertex_mapping_table[(*it2).second];
			}else{
				continue;
			}
			edge::iterator eit=g.adjacency_list[small].find(large);
			if (eit==g.adjacency_list[small].end())
				g.adjacency_list[small].insert(pair<vertex, weight>(large,1));
			else
				eit->second++;


			it1++;
			it2++;
		}
	}	
	
	vector<int> shrink_order;
	for(int i=0; i<V; i++)
		if (reverse_vertex_mapping_table[order[i]]< V-(int) isolated_vertices.size())
			shrink_order.push_back(order[i]);
	
	for (int i=0; i< V-(int) isolated_vertices.size(); i++)
		g.order[i]=reverse_vertex_mapping_table[shrink_order[i]];

	delete[] reverse_mapping;
	cout<<"Graph is created successfully"<<endl;
	return g;
}


Graph Hypergraph::Hypertograph_Loop(int* vertex_mapping_table, int* reverse_vertex_mapping_table, vector<int>& isolated_vertices){

	int* reverse_mapping;
	reverse_mapping=new int[V];
	for (int i=0; i<V; i++){
		reverse_mapping[order[i]]=i;
	}

	BitVec* sparse_matrix=new BitVec(V,1,V);
	
	//calculate sparse matrix
	for (int j=0; j<E; j++){
		hyperedge heg;
		for (int i=0; i<V; i++)
			if (transaction_matrix->get(j+1,i+1)==1){
				heg.insert(pair<ranking, vertex>(reverse_mapping[i],i));
			}
		if (heg.empty())
			continue;
		map<ranking,vertex>::iterator it1,it2;
		it1=heg.begin();
		it2=heg.begin();
		it2++;
		sparse_matrix->set_one((*it1).second+1, (*it1).second+1);
		while (it2!=heg.end()){
//			g.adjacency_matrix[(*it1).second][(*it2).second]++;//symmetric
//			g.adjacency_matrix[(*it2).second][(*it1).second]++;
			sparse_matrix->set_one((*it1).second+1, (*it2).second+1);
			sparse_matrix->set_one((*it2).second+1, (*it1).second+1);
			it1++;
			it2++;
		}
	}
	
	//eliminate isolated vertices
	int vertex_mapping_pointer=0;
	for (int i=0; i<V; i++){
		for (int j=0; j<V; j++){
			if (sparse_matrix->get(i+1,j+1)>0){
				vertex_mapping_table[vertex_mapping_pointer]=i;
				vertex_mapping_pointer++;
				break;
			}
			if (j==V-1)
				isolated_vertices.push_back(i);
		}
	}
	for (unsigned int i=0; i<isolated_vertices.size(); i++){
		vertex_mapping_table[vertex_mapping_pointer]=isolated_vertices[i];
		vertex_mapping_pointer++;
	}
	
	for (int i=0; i<V; i++)
		reverse_vertex_mapping_table[vertex_mapping_table[i]]=i;
	delete sparse_matrix;
	
	cout<<"V:"<<V<<endl;
	cout<<"compressed vertices:"<<isolated_vertices.size()<<endl;
	cout<<"delete sparse_matrix successfully"<<endl;

	//Create compressed graph
	Graph g(V-isolated_vertices.size());	
	for (int j=0; j<E; j++){
		hyperedge heg;
		for (int i=0; i<V; i++)
			if (transaction_matrix->get(j+1,i+1)==1){
				heg.insert(pair<ranking, vertex>(reverse_mapping[i],i));
			}
		if (heg.empty())
			continue;
		map<ranking,vertex>::iterator it1,it2;
		it1=heg.begin();
		it2=heg.begin();
		it2++;
		while (it2!=heg.end()){
//			g.adjacency_matrix[(*it1).second][(*it2).second]++;//symmetric
//			g.adjacency_matrix[(*it2).second][(*it1).second]++;
			int large=0;
			int small=0;
			if (reverse_vertex_mapping_table[(*it1).second]>reverse_vertex_mapping_table[(*it2).second]){
				small=reverse_vertex_mapping_table[(*it2).second];
				large=reverse_vertex_mapping_table[(*it1).second];
			}else if ((*it1).second<(*it2).second){
				small=reverse_vertex_mapping_table[(*it1).second];
				large=reverse_vertex_mapping_table[(*it2).second];
			}else{
				continue;
			}
			edge::iterator eit=g.adjacency_list[small].find(large);
			if (eit==g.adjacency_list[small].end())
				g.adjacency_list[small].insert(pair<vertex, weight>(large,1));
			else
				eit->second++;

			it1++;
			it2++;
		}
		it2--;
		it1=heg.begin();

//		g.adjacency_matrix[(*it1).second][(*it2).second]++;
//		g.adjacency_matrix[(*it2).second][(*it1).second]++;
		
		int large=0;
		int small=0;
		if (reverse_vertex_mapping_table[(*it1).second]>reverse_vertex_mapping_table[(*it2).second]){
			small=reverse_vertex_mapping_table[(*it2).second];
			large=reverse_vertex_mapping_table[(*it1).second];
		}else if ((*it1).second<(*it2).second){
			small=reverse_vertex_mapping_table[(*it1).second];
			large=reverse_vertex_mapping_table[(*it2).second];
		}else{
			continue;
		}
		edge::iterator eit=g.adjacency_list[small].find(large);
		if (eit==g.adjacency_list[small].end())
			g.adjacency_list[small].insert(pair<vertex, weight>(large,1));
		else
			eit->second++;		

	}

	vector<int> shrink_order;
	for(int i=0; i<V; i++)
		if (reverse_vertex_mapping_table[order[i]]< V-(int) isolated_vertices.size())
			shrink_order.push_back(order[i]);
	
	for (int i=0; i< V-(int) isolated_vertices.size(); i++)
		g.order[i]=reverse_vertex_mapping_table[shrink_order[i]];

	delete[] reverse_mapping;
	cout<<"Graph is created successfully"<<endl;
	return g;
}

/*
Graph Hypergraph::Hypertograph_Clique(){

	int* reverse_mapping;
	reverse_mapping=new int[V];
	for (int i=0; i<V; i++){
		reverse_mapping[order[i]]=i;
	}

	Graph g(V);

	for (int j=0; j<E; j++){
		hyperedge heg;
		for (int i=0; i<V; i++)
			if (transaction_matrix->get(j+1,i+1)==1){
				heg.insert(pair<ranking, vertex>(reverse_mapping[i],i));
			}
		if (heg.empty())
			continue;
		map<ranking,vertex>::iterator it;
		it=heg.begin();
		int* hyper_vertex_set;
		hyper_vertex_set=new int[heg.size()];
		int i=0;
		for (it=heg.begin();it!=heg.end();it++){
			hyper_vertex_set[i]=(*it).second;
			i++;
		}
		for (unsigned int i=0; i<heg.size(); i++)
			for (unsigned int j=i+1; j<heg.size(); j++){
				g.adjacency_matrix[hyper_vertex_set[i]][hyper_vertex_set[j]]++;
				g.adjacency_matrix[hyper_vertex_set[j]][hyper_vertex_set[i]]++;
			}
		delete[] hyper_vertex_set;

	}

	for (int i=0; i<V; i++){
		g.order[i]=order[i];
	}
	delete[] reverse_mapping;
	return g;
}
*/
Graph Hypergraph::Hypertograph_Multiloop(int* vertex_mapping_table, int* reverse_vertex_mapping_table, vector<int>& isolated_vertices){

	int* reverse_mapping;
	reverse_mapping=new int[V];
	for (int i=0; i<V; i++){
		reverse_mapping[order[i]]=i;
	}

	BitVec* sparse_matrix=new BitVec(V,1,V);
	
	//calculate sparse matrix
	for (int j=0; j<E; j++){
		hyperedge heg;
		for (int i=0; i<V; i++)
			if (transaction_matrix->get(j+1,i+1)==1){
				heg.insert(pair<ranking, vertex>(reverse_mapping[i],i));
			}
		if (heg.empty())
			continue;
		map<ranking,vertex>::iterator it1,it2;
		it1=heg.begin();
		it2=heg.begin();
		it2++;
		sparse_matrix->set_one((*it1).second+1, (*it1).second+1);
		while (it2!=heg.end()){
//			g.adjacency_matrix[(*it1).second][(*it2).second]++;//symmetric
//			g.adjacency_matrix[(*it2).second][(*it1).second]++;
			sparse_matrix->set_one((*it1).second+1, (*it2).second+1);
			sparse_matrix->set_one((*it2).second+1, (*it1).second+1);
			it1++;
			it2++;
		}
	}
	
	//eliminate isolated vertices
	int vertex_mapping_pointer=0;
	for (int i=0; i<V; i++){
		for (int j=0; j<V; j++){
			if (sparse_matrix->get(i+1,j+1)>0){
				vertex_mapping_table[vertex_mapping_pointer]=i;
				vertex_mapping_pointer++;
				break;
			}
			if (j==V-1)
				isolated_vertices.push_back(i);
		}
	}
	for (unsigned int i=0; i<isolated_vertices.size(); i++){
		vertex_mapping_table[vertex_mapping_pointer]=isolated_vertices[i];
		vertex_mapping_pointer++;
	}
	
	for (int i=0; i<V; i++)
		reverse_vertex_mapping_table[vertex_mapping_table[i]]=i;
	delete sparse_matrix;
	
	cout<<"V:"<<V<<endl;
	cout<<"compressed vertices:"<<isolated_vertices.size()<<endl;
	cout<<"delete sparse_matrix successfully"<<endl;
	
	//Create compressed graph
	Graph g(V-isolated_vertices.size());
	for (int j=0; j<E; j++){
		hyperedge heg;
		for (int i=0; i<V; i++)
			if (transaction_matrix->get(j+1,i+1)==1){
				heg.insert(pair<ranking, vertex>(reverse_mapping[i],i));
			}
		if (heg.empty())
			continue;
		map<ranking,vertex>::iterator it;
		it=heg.begin();
		int* hyper_vertex_set;
		hyper_vertex_set=new int[heg.size()];
		int i=0;
		for (it=heg.begin();it!=heg.end();it++){
			hyper_vertex_set[i]=(*it).second;
			i++;
		}
		for (unsigned int i=0; i<heg.size(); i++)
			for (unsigned int j=i+1; j<heg.size(); j++){
				int x=(int) pow(2,(double) heg.size()-j-2);//another seems working idea to reach round 3 is heg.size()-j-1
				int y=(int) pow(2,(double) i-1);//another seems working idea to reach round 3 is i-2
				x=max(1,x);
				y=max(1,y);
//				g.adjacency_matrix[hyper_vertex_set[i]][hyper_vertex_set[j]]=x*y;
//				g.adjacency_matrix[hyper_vertex_set[j]][hyper_vertex_set[i]]=x*y;

				int large=0;
				int small=0;
				if (reverse_vertex_mapping_table[hyper_vertex_set[i]]>reverse_vertex_mapping_table[hyper_vertex_set[j]]){
					small=reverse_vertex_mapping_table[hyper_vertex_set[j]];
					large=reverse_vertex_mapping_table[hyper_vertex_set[i]];
				}else if (reverse_vertex_mapping_table[hyper_vertex_set[i]]<reverse_vertex_mapping_table[hyper_vertex_set[j]]){
					small=reverse_vertex_mapping_table[hyper_vertex_set[i]];
					large=reverse_vertex_mapping_table[hyper_vertex_set[j]];
				}else{
					continue;
				}
		//		edge::iterator eit=g.adjacency_list[small].find(large);
		//		if (eit==g.adjacency_list[small].end())
				double normalize=(x*y)/pow(2,(double) heg.size());
				g.adjacency_list[small].insert(pair<vertex, weight>(large,max(normalize, numeric_limits<double>::min())));
		//		else
		//			eit->second++;

			
			}
		delete[] hyper_vertex_set;

	}
	
//	cout<<"min double:"<<numeric_limits<double>::min()<<endl;
 
	
	//g.Graph_display();
	
	vector<int> shrink_order;
	for(int i=0; i<V; i++)
		if (reverse_vertex_mapping_table[order[i]]< V-(int) isolated_vertices.size())
			shrink_order.push_back(order[i]);
/*	
	cout<<"shrink order:";
	for (vector<int>::iterator it=shrink_order.begin(); it!=shrink_order.end();it++)
		cout<<*it<<",";
	cout<<endl;
	cout<<"here:";
*/	
	for (int i=0; i< V-(int) isolated_vertices.size(); i++){
		g.order[i]=reverse_vertex_mapping_table[shrink_order[i]];
//		cout<<g.order[i]<<",";
	}
//	cout<<endl;

	delete[] reverse_mapping;
	
	cout<<"Graph is created successfully"<<endl;
	return g;

}


void Hypergraph::Hypergraph_display(){
	cout<<"Hypergraph is displayed as follow:"<<endl;
	cout<<"E:"<<E<<";";
	cout<<"V:"<<V;
	cout<<endl;
	for (int i=0; i<E; i++){
		cout<<i<<":";
		for (int j=0; j<V; j++)
			if (transaction_matrix->get(i+1,j+1)==1)
				cout<<j<<",";
		cout<<endl;
	}
}





