#include <assert.h>
#include <limits>
#include <stdlib.h>
#include <vector>
#include "BitVec/BitVec.hh"
#include "types.hh"
#include "write_image.hh"

int main(int argc, char **argv)
{
  if (argc < 2)
  {
    fprintf(stderr, "Usage: %s filename\n", argv[0]);
    exit(1);
  }

  FILE *f;
  if ((f = fopen(argv[1], "rb")) == NULL)
  {
    perror("fopen");
    exit(2);
  }

  char *line = NULL;
  size_t line_size = 0;
  trans_t max_trans = 0;
  item_t min_item = std::numeric_limits<item_t>::max();
  item_t max_item = std::numeric_limits<item_t>::min();

  char *line_scan;
  while (getline(&line, &line_size, f) != -1)
  {
    line_scan = line;

    item_t item;
    int item_str_len;
    while (sscanf(line_scan, "%d%n", &item, &item_str_len) == 1)
    {
      min_item = std::min(min_item, item);
      max_item = std::max(max_item, item);
      line_scan += item_str_len;
    }

    assert(line_scan[0] == ';');
    ++line_scan;

    trans_t trans;
    int trans_str_len;
    while (sscanf(line_scan, "%d%n", &trans, &trans_str_len) == 1)
    {
      max_trans = std::max(max_trans, trans);
      line_scan += trans_str_len;
    }
  }

  printf("min_item %d\n", min_item);
  printf("max_item %d\n", max_item);
  printf("max_trans %d\n", max_trans);

  rewind(f);

  BitVec bv(max_trans, min_item, max_item);

  while (getline(&line, &line_size, f) != -1)
  {
    line_scan = line;
    
    std::vector<item_t> items;

    item_t item;
    int item_str_len;
    while (sscanf(line_scan, "%d%n", &item, &item_str_len) == 1)
    {
      items.push_back(item);
      line_scan += item_str_len;
    }

    assert(line_scan[0] == ';');
    ++line_scan;

    std::vector<trans_t> transactions;

    trans_t trans;
    int trans_str_len;
    while (sscanf(line_scan, "%d%n", &trans, &trans_str_len) == 1)
    {
      transactions.push_back(trans);
      line_scan += trans_str_len;
    }

    for (std::vector<trans_t>::size_type t=0; t<transactions.size(); ++t)
    {
      for (std::vector<item_t>::size_type i=0; i<items.size(); ++i)
      {
        bv.set_one(transactions[t], items[i]);
      }
    }
  }

  if (fclose(f) != 0)
  {
    perror("fclose");
    exit(3);
  }

  //std::cout << bv;

  write_image3(bv, std::string(argv[1]));

  return 0;
}

