#include <algorithm>
#include <assert.h>
#include <errno.h>
#include <limits>
#include <stdlib.h>
#include <string.h>
#include <vector>
#include "BitVec/BitVec.hh"
#include "types.hh"

int main(int argc, char **argv)
{
  if (argc < 4)
  {
    fprintf(stderr, "Usage: %s trans_db_file hyper_file num_hyperrec\n", argv[0]);
    exit(1);
  }

  errno = 0;
  long int num_hyperrec = strtol(argv[3], NULL, 10);
  if (errno != 0)
  {
    perror("converting num_hyperrec parameter");
    exit(5);
  }

  FILE *ds_f;
  if ((ds_f = fopen(argv[1], "rb")) == NULL)
  {
    perror("fopen trans_db file");
    exit(1);
  }

  char *line = NULL;
  size_t line_size = 0;

  item_t min_item = std::numeric_limits<item_t>::max();
  item_t max_item = std::numeric_limits<item_t>::min();

  int line_ct = 0;
  char *line_scan;
  while (getline(&line, &line_size, ds_f) != -1)
  {
    line_scan = line;
    ++line_ct;

    item_t item;
    int item_str_len;
    while (sscanf(line_scan, "%d%n", &item, &item_str_len) == 1)
    {
      min_item = std::min(min_item, item);
      max_item = std::max(max_item, item);
      line_scan += item_str_len;
    }
  }

  const trans_t max_trans = line_ct;

  printf("min_item %d\n", min_item);
  printf("max_item %d\n", max_item);
  printf("max_trans %d\n", max_trans);

  // items only; row axis is a sequential counter 
  BitVec item_matrix(num_hyperrec, min_item, max_item); 

  // transactions only; row axis is a sequential counter 
  BitVec trans_matrix(num_hyperrec, 1, max_trans); 

  FILE *hr_f;
  if ((hr_f = fopen(argv[2], "rb")) == NULL)
  {
    perror("fopen");
    exit(2);
  }

  line_ct = 0;
  while (getline(&line, &line_size, hr_f) != -1 && line_ct < num_hyperrec)
  {
    line_scan = line;
    ++line_ct;

    item_t item;
    int item_str_len;
    while (sscanf(line_scan, "%d%n", &item, &item_str_len) == 1)
    {
      item_matrix.set_one(line_ct, item);
      line_scan += item_str_len;
    }

    assert(line_scan[0] == ';');
    ++line_scan;

    trans_t trans;
    int trans_str_len;
    while (sscanf(line_scan, "%d%n", &trans, &trans_str_len) == 1)
    {
      trans_matrix.set_one(line_ct, trans);
      line_scan += trans_str_len;
    }

  }

  if (fclose(hr_f) != 0)
  {
    perror("fclose");
    exit(3);
  }
  hr_f = NULL;

  // original database; row axis is transactions, column axis is items
  BitVec trans_db(max_trans, min_item, max_item);

  rewind(ds_f);

  line_ct = 0;
  while (getline(&line, &line_size, ds_f) != -1)
  {
    line_scan = line;
    ++line_ct;

    item_t item;
    int item_str_len;
    while (sscanf(line_scan, "%d%n", &item, &item_str_len) == 1)
    {
      trans_db.set_one(line_ct, item);

      line_scan += item_str_len;
    }
  }

  if (fclose(ds_f) != 0)
  {
    perror("fclose");
    exit(4);
  }

  std::cout << "items matrix:\n" << item_matrix << "\n";

  getchar();
  
  std::cout << "transactions matrix:\n" << trans_matrix << "\n";

  getchar();
  
  std::cout << "original database:\n" << trans_db << "\n";

  return 0;
}

