#include "Graph.h"
#include <assert.h>
#include <cstdlib>
#include <iomanip>
#include <limits>
#include <sys/time.h>
#include <stdlib.h>
#include <stdio.h>
#include <stack>
#include <string.h>
#include <set>
#include "write_image.hh"
#include "HyperRectBounds.hh"

//#define DEBUG

enum SolutionType { PATH, CYCLE, MAXVALIDCYCLES };

uint64_t hyperorder_usec = 0;
uint64_t minla_usec = 0;

std::string solution_type_str(SolutionType st)
{
  std::string st_str;
  switch (st)
  {
  case PATH:
    st_str = "path";
    break;
  case CYCLE:
    st_str = "cycle";
    break;
  case MAXVALIDCYCLES:
    st_str = "maxvalidcycles";
    break;
  }
  return st_str;
}
  
void enum_orders(Graph *g, int depth, int *best_order, int *lowest_cost, bool *used_vertices){
	if (depth==g->sizeV()){
		if (g->Graphcost()<*lowest_cost){
			*lowest_cost=g->Graphcost();
			for (int j=0; j<g->sizeV(); j++){
				best_order[j]=g->order[j];
//				cout<<g->order[j]<<endl;
			}
		}
		return;
	}


	int i=0;
	while (i<g->sizeV()){
		if(used_vertices[i]==false){
			g->order[depth]=i;
			used_vertices[i]=true;
			enum_orders(g, depth+1, best_order, lowest_cost, used_vertices);
			used_vertices[i]=false;
		}

		i++;
	}

}

void enum_orders_hg(Hypergraph *hg, int depth, int *best_order, int *lowest_cost, bool *used_vertices){
	if (depth==hg->sizeV()){
		if (hg->Hypercost()<*lowest_cost){
			*lowest_cost=hg->Hypercost();
			for (int j=0; j<hg->sizeV(); j++){
				best_order[j]=hg->order[j];
//				cout<<g->order[j]<<endl;
			}
		}
		return;
	}


	int i=0;
	while (i<hg->sizeV()){
		if(used_vertices[i]==false){
			hg->order[depth]=i;
			used_vertices[i]=true;
			enum_orders_hg(hg, depth+1, best_order, lowest_cost, used_vertices);
			used_vertices[i]=false;
		}

		i++;
	}

}

void hypergraph_generator(Hypergraph& hg){
	for (int i=0; i<hg.sizeE(); i++)
		for (int j=0; j<hg.sizeV(); j++){
			//(8,8)135 works well for line and loop
			//(8,8)131 works for multiloop to reach round 3
			srand(i*168+j);
			int k=rand();
	//		cout<<"k:"<<k<<endl;
			if (k % 100 < 30)
				hg.transaction_matrix->set_one(i+1,j+1);
			else
				hg.transaction_matrix->set_zero(i+1,j+1);
		}

}

//int main_original(int argc, char* argv[]) {
void magic_sorting(Hypergraph* hg, BitVec &trans_db, const BitVec trans_db_image, string &dataset_name, SolutionType ST){



//	std::cout << "database matrix before reorder1:\n" << trans_db << "\n";	
	#ifdef DEBUG
//	cout<<"trans_db at the beginning of magic sorting:"<<endl;
//	cout<<"trans_db (3,30)"<<trans_db.get(3,30)<<endl;
//	cout<<trans_db_image;
	#endif

	#ifdef DEBUG
//	cout<<"trans_db_image at the beginning of magic sorting:"<<endl;
//	cout<<"trans_db_image (3,30)"<<trans_db_image.get(3,30)<<endl;
//	cout<<trans_db_image;
	#endif
	

	bool* used_vertices_for_search;
	used_vertices_for_search=new bool [hg->sizeV()];
	
	int* hyper_initial_order;
	hyper_initial_order=new int [hg->sizeV()];
	for (int i=0; i<hg->sizeV(); i++){
		used_vertices_for_search[i]=false;
		hyper_initial_order[i]=i;
	}

//	stack<int> mystack;
//	mystack.push(0);//the first vertex is put into the stack;
//	used_vertices_for_search[0]=true;
//	hyper_initial_order[0]=0;
//	bool pop_stack_tag=false;

/*
	do{//try different initial orders
		if (mystack.size()<hg->sizeV()){
			if (pop_stack_tag==false){
				for (int i=0; i<hg->sizeV(); i++)
					if(used_vertices_for_search[i]==false){
						hyper_initial_order[mystack.size()]=i;
						used_vertices_for_search[i]=true;
						mystack.push(i);
						break;
					}
			}
			else{//pop_stack_tag==true
				int current_v=mystack.top();
				mystack.pop();
				used_vertices_for_search[current_v]=false;
				for (int i=current_v+1; i<hg->sizeV(); i++)
					if(used_vertices_for_search[i]==false){
						hyper_initial_order[mystack.size()]=i;
						used_vertices_for_search[i]=true;
						mystack.push(i);
						pop_stack_tag=false;
						break;
					}
			}

			continue;
		}
*/
		#ifdef DEBUG
		cout<<"hypergraph initial order:"<<endl;
		#endif
		for (int i=0; i<hg->sizeV(); i++){
			hg->order[i]=hyper_initial_order[i];
			#ifdef DEBUG
			cout<<hyper_initial_order[i]<<",";
			#endif
		}

		#ifdef DEBUG
		cout<<endl;
		#endif





		int Hypergraph_cost=hg->Hypercost();

		#ifdef DEBUG
		cout <<"Hyper graph initial cost:"<< Hypergraph_cost<< endl;
		cout<<endl;
		#endif


		//repeat the following until no improvement:
		int round=1;
		do{

			#ifdef DEBUG
			cout<<"round:"<<round<<endl;
			#endif

			round++;
			Hypergraph_cost=hg->Hypercost();

			#ifdef DEBUG
//			hg->Hypergraph_display();
			cout<<"Hyper graph cost:"<<hg->Hypercost()<<endl;
			#endif

			    

			
			int* vertex_mapping_table; //to eliminate isolated vertices
			int* reverse_vertex_mapping_table; 
			vertex_mapping_table=new int [hg->sizeV()];
			reverse_vertex_mapping_table=new int[hg->sizeV()];
			vector<int> isolated_vertices;
			
			//convert the hypergraph into a regular graph
			Graph g;
//			Graph g=hg->Hypertograph_Line(vertex_mapping_table, reverse_vertex_mapping_table, isolated_vertices);
			if (ST==PATH)
				g=hg->Hypertograph_Line(vertex_mapping_table, reverse_vertex_mapping_table, isolated_vertices);
			else if (ST==CYCLE)
				g=hg->Hypertograph_Loop(vertex_mapping_table, reverse_vertex_mapping_table, isolated_vertices);
			//Graph g=hg->Hypertograph_Clique();
			else if (ST==MAXVALIDCYCLES)
				g=hg->Hypertograph_Multiloop(vertex_mapping_table, reverse_vertex_mapping_table, isolated_vertices);
			
			#ifdef DEBUG
			cout<<"vertex mapping table:";
//			getchar();
			for (int i=0; i<hg->sizeV(); i++)
				cout<<vertex_mapping_table[i]<<",";
			cout<<endl;
//			getchar();
			
			cout<<"vertex reverse mapping table:";
			for (int i=0; i<hg->sizeV(); i++)
				cout<<reverse_vertex_mapping_table[i]<<",";
			cout<<endl;
//			getchar();

			cout<<"isolated_vertices:";
			for (size_t i=0; i<isolated_vertices.size(); i++)
				cout<<isolated_vertices[i]<<",";
			cout<<endl;			
			#endif
			#ifdef DEBUG
			cout<<"graph V:"<<g.sizeV()<<endl;
			cout<<"graph E:"<<g.sizeE()<<endl;
//			g.Graph_display();
			cout<<"number of vertices in g:"<<g.sizeV()<<endl;
			cout<<"number of compressed vertices:"<<isolated_vertices.size()<<endl;
			#endif
			
			#ifdef DEBUG
			cout <<"Graph order:";
			for (int i=0; i<g.sizeV(); i++)
				cout<<g.order[i]<<",";
			cout<<endl;
			cout <<"Graph cost:"<< g.Graphcost()<< endl;
			g.Graph_display();
			cout<<endl;
			#endif
			
			//starting calling minla with connected component
			set<int> visited_vertices;
			set<int> unvisited_vertices;
			set<int> running_stack;

			
			
			for (int i=0; i<g.sizeV(); i++)
				unvisited_vertices.insert(i);
				
			int* best_order;
			best_order=new int[g.sizeV()];
			int order_offset=0;
			
			
			while (unvisited_vertices.size()>0){
				set<int> connected_component;
				set<int>::iterator root=unvisited_vertices.begin();
				visited_vertices.insert(*root);
				running_stack.insert(*root);
				connected_component.insert(*root);
				unvisited_vertices.erase(root);
				while(running_stack.size()>0){
					set<int>::iterator it_current_vertex=running_stack.begin();
					int current_vertex_id=*it_current_vertex;
					running_stack.erase(current_vertex_id);
					unvisited_vertices.erase(current_vertex_id);
					visited_vertices.insert(current_vertex_id);

					for(edge::iterator it=g.adjacency_list[current_vertex_id].begin(); it!=g.adjacency_list[current_vertex_id].end(); it++){
						if (visited_vertices.find((*it).first)==visited_vertices.end()){
							running_stack.insert((*it).first);
							visited_vertices.insert((*it).first);
							connected_component.insert((*it).first);
							unvisited_vertices.erase((*it).first);
						}
					}
					for (int i=0; i<g.sizeV(); i++){
						if ((g.adjacency_list[i].find(current_vertex_id)!=g.adjacency_list[i].end())&&(visited_vertices.find(i)==visited_vertices.end())){
							running_stack.insert(i);
							visited_vertices.insert(i);
							connected_component.insert(i);
							unvisited_vertices.erase(i);
						}
					}

					
				}
				
				int *connected_mapping;
				int *reverse_connected_mapping;
				int connected_edges=0;
				connected_mapping=new int[connected_component.size()];
				reverse_connected_mapping=new int[g.sizeV()];//the maximum value is g.sizeV();
				int ii=0;
				#ifdef DEBUG
				cout<<"connected mapping table:"<<endl;
				#endif
				for (set<int>::iterator it=connected_component.begin(); it!=connected_component.end(); it++, ii++){
					connected_mapping[ii]=*it;
					#ifdef DEBUG
					cout<<*it<<",";
					#endif
					connected_edges+=g.adjacency_list[*it].size();
				}
				#ifdef DEBUG
				cout<<endl;
				#endif
				
				
				for(int i=0; i<(int)connected_component.size(); i++){
					reverse_connected_mapping[connected_mapping[i]]=i;//many values are empty;
				}
				
				#ifdef DEBUG
				cout<<"reverse_connected_mapping table:"<<endl;
				for(int i=0; i<(int)g.sizeV(); i++){
					cout<<reverse_connected_mapping[i]<<",";
				}
				cout<<endl;
				#endif
			
				string out_filename(dataset_name+".rmf");
				ofstream outfile(out_filename.c_str());
				if (!outfile) {
					cout << "Error: Cannot create *.rmf" << endl;
					exit(1);
				}


			
				outfile<<"p ghct "<<connected_component.size()<<" "<<connected_edges<<endl;

				for (set<int>::iterator it=connected_component.begin(); it!=connected_component.end(); it++)
					for (edge::iterator it1=(g.adjacency_list[*it]).begin(); it1!=(g.adjacency_list[*it]).end(); it1++){
						outfile<<"e "<<reverse_connected_mapping[*it]+1<<" "<<reverse_connected_mapping[it1->first]+1<<" "<<it1->second<<endl;			
				}
			
				outfile.close();
				
				struct timeval tv;
				if (gettimeofday(&tv, NULL) == -1)
				{
				  perror("gettimeofday");
				  exit(1);
				}
				uint64_t timing_minla_start = (uint64_t)tv.tv_sec * (uint64_t)1000000 + (uint64_t)tv.tv_usec;
			
				string command("bin/minla "+out_filename+" minla_params");
				cout << "issuing system call: " << command << "\n";
				int i=system(command.c_str());
				if (WEXITSTATUS(i)!=0){
					cout<< "Error issuing system call to minla. Exit status is " << WEXITSTATUS(i) << ".  Terminating program.\n";
					exit(1);
				}

				if (gettimeofday(&tv, NULL) == -1)
				{
				  perror("gettimeofday");
				  exit(1);
				}

				minla_usec += ((uint64_t)tv.tv_sec * (uint64_t)1000000 + (uint64_t)tv.tv_usec) - timing_minla_start;
			
				int* partial_best_order;
				partial_best_order=new int [connected_component.size()];



				string in_filename(out_filename+".order");
				ifstream infile(in_filename.c_str());
				if (!infile) {
					perror("");
					cout << "Error: Cannot open " << in_filename << " for reading." << endl;
					exit(1);
				}
			
				#ifdef DEBUG
				cout <<"Graph order generated by MinLa:";
				#endif
				for (int i=0; i<(int)connected_component.size(); i++){
					infile>>partial_best_order[i];
					#ifdef DEBUG
					cout<<partial_best_order[i]<<",";
					#endif
				}
				#ifdef DEBUG
				cout<<endl;
				#endif
				
				for (int i=0; i<(int)connected_component.size(); i++){
					best_order[order_offset]=connected_mapping[partial_best_order[i]-1];
					order_offset++;
				}
				
				infile.close();
				if(remove(in_filename.c_str())!=0){
					cout<< "Delete file error"<<endl;
					exit(1);
				}
				
				delete connected_mapping;
				delete reverse_connected_mapping;
				delete partial_best_order;
				
				
			}
			
			#ifdef DEBUG
			cout <<"Graph improved order transalated from MinLa:";
			#endif
			
			for (int i=0; i<g.sizeV(); i++){
//				g.order[i]=best_order[i]-1;//no connected component version
				g.order[i]=best_order[i];
//				hg->order[i]=vertex_mapping_table[best_order[i]-1];//no connected component version
				hg->order[i]=vertex_mapping_table[best_order[i]];
				#ifdef DEBUG
				cout<<hg->order[i]<<",";
				#endif
			}
			
			for (int i=hg->sizeV()-isolated_vertices.size(); i<hg->sizeV(); i++){
				hg->order[i]=vertex_mapping_table[i];
				#ifdef DEBUG
				cout<<hg->order[i]<<",";
				#endif
			}
			#ifdef DEBUG
			cout<<endl;
			#endif

			

			


			
			#ifdef DEBUG
			cout <<"Graph improved cost:"<< g.Graphcost()<< endl;
			#endif


			#ifdef DEBUG
			cout<<endl;
			#endif

			delete[] vertex_mapping_table;
			delete[] reverse_vertex_mapping_table;			
//			delete in_filename;
//			delete out_filename;
			delete[] best_order;

				

		}while (hg->Hypercost()<Hypergraph_cost);
					
		cout <<"Hypergraph final cost:"<< hg->Hypercost() << endl;


/*
		if (round>=4){
			cout<<"found it!!!"<<endl;
			break;
		}



		int current_v=mystack.top();
		mystack.pop();
		used_vertices_for_search[current_v]=false;
		pop_stack_tag=true;
		
		#ifdef DEBUG
		cout<<endl;
		cout<<endl;
		cout<<endl;
		#endif

	}while (mystack.size()>0);
*/

/*//calculate optimal order for Hypergraph
	int* best_order;
	best_order=new int[hg->sizeV()];
	int lowest_cost=MAX_VAL;
	bool* used_vertices;
	used_vertices=new bool[hg->sizeV()]; 
	for (int i=0; i<hg->sizeV(); i++)
		used_vertices[i]=false;
	enum_orders_hg(&hg,0,best_order,&lowest_cost, used_vertices);
	cout <<"Hypergraph optimized cost:"<< lowest_cost<< endl;
	cout <<"Hypergraph optimized order:";
	for (int i=0; i<hg->sizeV(); i++){
		cout<<best_order[i]<<",";
	}
	cout<<endl;
	delete[] best_order;
	delete[] used_vertices;
*/	
	int* reverse_final_order;
	reverse_final_order=new int[hg->sizeV()];
	for (int i=0; i<hg->sizeV(); i++){
		reverse_final_order[hg->order[i]]=i;
	}

	#ifdef DEBUG
	cout<<"reverse_final_order:";
	for (int i=0; i<hg->sizeV(); i++)
		cout<<reverse_final_order[i]<<",";
	cout<<endl;
	#endif

	#ifdef DEBUG	
//	std::cout << "database matrix before reorder2:\n" << trans_db << "\n";	
	#endif
	
	#ifdef DEBUG
//	cout<<"trans_db_image:"<<endl;
//	cout<<"trans_db_image (3,30)"<<trans_db_image.get(3,30)<<endl;
//	cout<<trans_db_image;
	#endif
	
	for (int i=0; i<(int) trans_db.num_trans(); i++)
		for (int j=0; j<(int) trans_db.max_item(); j++){
			if ( trans_db_image.get(i+1,j+1)==0)
				trans_db.set_zero(i+1,reverse_final_order[j]+1);
			else
				trans_db.set_one(i+1,reverse_final_order[j]+1);
		}
	
	#ifdef DEBUG
//	cout<<"trans_db in magic sorting:"<<endl;
//	cout<<trans_db;
	#endif

	#ifdef DEBUG		
//	 std::cout << "database matrix after reorder:\n" << trans_db << "\n";
	#endif

	//release memory

	delete[] reverse_final_order;

	delete[] used_vertices_for_search;

	delete[] hyper_initial_order;
}//end of magic_sorting

int main1(int argc, char **argv)
{
	cout<<"test code start."<<endl;
	BitVec* sparse_matrix=new BitVec(77512,1,77512);
	
//	for (int i=0; i<77512; i++)
//		for (int j=0; j<77512; j++)
//			sparse_matrix->set_zero(i+1,j+1);
	
	sparse_matrix->set_one(55407,55615);
	sparse_matrix->set_one(55615,55407);
	cout<<"(204,28030)"<<sparse_matrix->get(205,28031)<<endl;
	delete sparse_matrix;
	
	return 0;
	


}

int main(int argc, char **argv)
{

  if (argc < 4)
  {
    fprintf(stderr, "Usage: %s trans_db_file hyper_file num_hyperrec [-s path|cycle|maxvalidcycles] [-z full|zoomed]\n", argv[0]);
    exit(1);
  }


  SolutionType solution_type = PATH; // default solution type
  bool zoom = true; // default whether-to-zoom

  int cur_opt;
  optind = 4;
  while ((cur_opt = getopt(argc, argv, "s:z:")) != -1)
  {
    switch(cur_opt)
    {
    case 's':
      if (strcmp(optarg, "path") == 0)
      {
        solution_type = PATH;
      } else if (strcmp(optarg, "cycle") == 0) {
        solution_type = CYCLE;
      } else if (strcmp(optarg, "maxvalidcycles") == 0) {
        solution_type = MAXVALIDCYCLES;
      } else {
        fprintf(stderr, "Error: Invalid solution type %s.\n", optarg);
        exit(1);
      }
      break;
    case 'z':
      if (strcmp(optarg, "full") == 0)
      {
        zoom = false;
      } else if (strcmp(optarg, "zoomed") == 0) {
        zoom = true;
      } else {
        fprintf(stderr, "Error: invalid zoom value %s\n", optarg);
	exit(1);
      }
      break;
    }
  }

  char *endptr;
  errno = 0;
  long int num_hyperrec = strtol(argv[3], &endptr, 10);
  if (errno != 0)
  {
    perror("converting num_hyperrec parameter");
    exit(5);
  }
  if (endptr == argv[3])
  {
    fprintf(stderr, "Error converting numhyperrec param: no digits found\n");
    exit(5);
  }

  FILE *ds_f;
  if ((ds_f = fopen(argv[1], "rb")) == NULL)
  {
    perror("fopen trans_db file");
    exit(1);
  }

  char *line = NULL;
  size_t line_size = 0;

  item_t min_item = std::numeric_limits<item_t>::max();
  item_t max_item = std::numeric_limits<item_t>::min();

  int line_ct = 0;
  char *line_scan;
  while (getline(&line, &line_size, ds_f) != -1)
  {
    line_scan = line;
    ++line_ct;

    item_t item;
    int item_str_len;
    while (sscanf(line_scan, "%d%n", &item, &item_str_len) == 1)
    {
      min_item = std::min(min_item, item);
      max_item = std::max(max_item, item);
      line_scan += item_str_len;
    }
  }

  const trans_t max_trans = line_ct;

  printf("min_item %d\n", min_item);
  printf("max_item %d\n", max_item);
  printf("max_trans %d\n", max_trans);

  // items only; row axis is a sequential counter 
  BitVec item_matrix(num_hyperrec, 1, max_item-min_item+1); 

  // transactions only; row axis is a sequential counter 
  BitVec trans_matrix(num_hyperrec, 1, max_trans);

  FILE *hr_f;
  if ((hr_f = fopen(argv[2], "rb")) == NULL)
  {
    perror("fopen");
    exit(2);
  }

  std::vector<std::vector<int> > hr_items;
  std::vector<std::vector<int> > hr_trans;

  line_ct = 0;
  while (getline(&line, &line_size, hr_f) != -1)
  {
    line_scan = line;
    ++line_ct;

    hr_items.push_back(std::vector<int>());
    item_t item;
    int item_str_len;
    while (sscanf(line_scan, "%d%n", &item, &item_str_len) == 1)
    {
      //std::cout << "item_matrix.set_one(" << line_ct << ", " << item-min_item+1 << ")\n";
      //item_matrix.set_one(line_ct, item-min_item+1);
      if (line_ct <= num_hyperrec)
      {
 //       item_matrix.set_one(line_ct, item);
       item_matrix.set_one(line_ct, item-min_item+1);
      }
      hr_items[line_ct - 1].push_back(item-min_item+1);
      line_scan += item_str_len;
    }

    assert(line_scan[0] == ';');
    ++line_scan;

    hr_trans.push_back(std::vector<int>());
    trans_t trans;
    int trans_str_len;
    while (sscanf(line_scan, "%d%n", &trans, &trans_str_len) == 1)
    {
      if (line_ct <= num_hyperrec)
      {
        trans_matrix.set_one(line_ct, trans);
      }
      hr_trans[line_ct - 1].push_back(trans);
      line_scan += trans_str_len;
    }

  }

  if (fclose(hr_f) != 0)
  {
    perror("fclose");
    exit(3);
  }
  hr_f = NULL;

/*
//std::cout << "items matrix:\n" << item_matrix << "\n";

std::cout << "item_matrix.get(1, 1): " << item_matrix.get(1, 1) << "\n";
std::cout << "item_matrix.get(1, 0): " << item_matrix.get(1, 0) << "\n";
std::cout << "item_matrix.get(0, 1): " << item_matrix.get(0, 1) << "\n";

std::cout << "item_matrix.get(1, 8): " << item_matrix.get(1, 8) << "\n";
std::cout << "item_matrix.get(1, 9): " << item_matrix.get(1, 9) << "\n";

std::cout << "----------\n";

item_matrix.set_one(1, 1);
std::cout << "item_matrix.get(1, 1): " << item_matrix.get(1, 1) << "\n";

exit(1);
*/


  // original database; row axis is transactions, column axis is items
  BitVec trans_db(max_trans, 1, max_item-min_item+1);

  rewind(ds_f);

  line_ct = 0;
  while (getline(&line, &line_size, ds_f) != -1)
  {
    line_scan = line;
    ++line_ct;

    item_t item;
    int item_str_len;
    while (sscanf(line_scan, "%d%n", &item, &item_str_len) == 1)
    {
      trans_db.set_one(line_ct, item-min_item+1);
      line_scan += item_str_len;
    }
  }

  if (fclose(ds_f) != 0)
  {
    perror("fclose");
    exit(4);
  }

// getchar();
  
// std::cout << "transactions matrix:\n" << trans_matrix << "\n";

//  getchar();
  
 // std::cout << "original database:\n" << trans_db << "\n";


  
  string dataset_name(argv[1]);
  size_t pos=dataset_name.find_last_of("/");
  dataset_name=dataset_name.substr(pos+1);
  
  #ifdef DEBUG
  cout<<dataset_name<<endl;
  #endif

  //std::string orig_fname(std::string("images/") + dataset_name.substr(0, dataset_name.length() - 4) + ".unordered");
  std::string orig_fname(dataset_name.substr(0, dataset_name.length() - 4) + ".unordered");
  //cout << "writing orig output to file " << orig_fname << "\n";
  //trans_db.write_to_trans_file(orig_fname);
  std::cout << "writing original transaction db image " << orig_fname << "\n";
  write_image3(trans_db, orig_fname);

//  cout<<"trans_db test"<<endl;
// cout<<"trans_db (1,34):"<<trans_db.get(1,34)<<endl;
//  getchar();

  
  struct timeval tv;
  if (gettimeofday(&tv, NULL) == -1)
  {
    perror("gettimeofday");
    exit(1);
  }
  uint64_t timing_hyperorder_start = (uint64_t)tv.tv_sec * (uint64_t)1000000 + (uint64_t)tv.tv_usec;
  
  Hypergraph* hg=new Hypergraph(trans_db.max_item()-trans_db.min_item()+1, trans_db.num_trans());
  
 #ifdef DEBUG
// cout<<"setting hg->transaction_matrix:"<<endl;
// cout<<item_matrix<<endl;
 #endif
 for (int i=0; i<(int) item_matrix.num_trans(); i++)
	for (int j=0; j<(int)item_matrix.max_item()-(int)item_matrix.min_item()+1; j++){
		if (item_matrix.get(i+1,j+1)==1){
			#ifdef DEBUG
			cout<<"("<<i<<","<<j<<"),";
			#endif
			hg->transaction_matrix->set_one(i+1,j+1);
		}
		//else
			//hg->transaction_matrix->set_zero(i+1,j+1);
	}
#ifdef DEBUG
cout<<endl;
#endif
	

//	std::cout << "database matrix before reorder0:\n" << trans_db << "\n";	
	magic_sorting(hg, trans_db, trans_db, dataset_name, solution_type);

	std::vector<int> item_order_idx(hg->sizeV());
        cout << "item order: ";
        for (int i=0; i<hg->sizeV(); ++i)
 	{
	  if (i != 0) cout << ", ";
	  const int to = hg->order[i] + 1;
          cout << to;
	  item_order_idx.at(to - 1) = i;
	}
	cout << "\n";

	delete hg;
	hg = NULL;
	
	cout<<"delete successful"<<endl;

	#ifdef DEBUG
//	cout<<"half-final trans_db before transpose:"<<endl;
//	cout<<trans_db;
	#endif

	trans_db.transpose();

	#ifdef DEBUG
//	cout<<"half-final trans_db:"<<endl;
//	cout<<trans_db;
	#endif
 
	cout<<"transpose successful"<<endl;
 #ifdef DEBUG
 cout<<"setting hg1->transaction_matrix:"<<endl;
 cout<<trans_matrix<<endl;
 #endif

	Hypergraph* hg1=new Hypergraph(trans_db.max_item()-trans_db.min_item()+1, trans_db.num_trans());
  
	cout<<"create new hg1 successful"<<endl;
	for (int i=0; i<(int) trans_matrix.num_trans(); i++)
		for (int j=0; j<(int) trans_matrix.max_item()-(int)trans_matrix.min_item()+1; j++){
			if (trans_matrix.get(i+1,j+1)==1)
				hg1->transaction_matrix->set_one(i+1,j+1);
			//else
				//hg1->transaction_matrix->set_zero(i+1,j+1);
	}

	
	cout<<"initializng hg1 successful"<<endl; 
	magic_sorting(hg1, trans_db, trans_db, dataset_name, solution_type);
	cout<<"sort successful"<<endl;

	std::vector<int> trans_order_idx(hg1->sizeV());
	cout << "transaction order: ";
	for (int i=0; i<hg1->sizeV(); ++i)
	{
	  if (i != 0) cout << ", ";
	  const int to = hg1->order[i] + 1;
	  cout << to;
	  trans_order_idx.at(to - 1) = i;
	}
	cout << "\n";
	
	delete hg1;

	trans_db.transpose();

	#ifdef DEBUG
//	cout<<"final trans_db:"<<endl;
//	cout<<trans_db;
	#endif
	

        if (gettimeofday(&tv, NULL) == -1)
        {
          perror("gettimeofday");
          exit(1);
        }

	hyperorder_usec = ((uint64_t)tv.tv_sec * (uint64_t)1000000 + (uint64_t)tv.tv_usec) - timing_hyperorder_start - minla_usec;
	
	cout << "minla time (sec): " << (double)minla_usec / exp10(6) << "\n";
	cout << "hyperorder time (sec): " << (double)hyperorder_usec / exp10(6) << "\n";


	std::vector<HyperRectBounds> hyperrects;
	HyperRectBounds* render_area = new HyperRectBounds();
	assert(hr_items.size() == hr_trans.size());
	for (int i=0; i<(int)hr_items.size(); ++i)
	{
	  HyperRectBounds hrb;
	  for (size_t j=0; j<hr_items[i].size(); ++j)
          {
            const int cur_hr_item = hr_items[i][j];
	    const int ioi = item_order_idx.at(cur_hr_item - 1) + 1;
            //cout << "mapping cur_hr_item " << cur_hr_item << " to order index " << ioi << "\n";
	    hrb.minx = std::min(hrb.minx, ioi);
            hrb.maxx = std::max(hrb.maxx, ioi);
	  }
          
	  //std::cout << "trans " << i << ": ";
          for (size_t j=0; j<hr_trans[i].size(); ++j)
	  {
	    //std::cout << "( " <<  hr_trans[i][j] << ", " << trans_order_idx[hr_trans[i][j] -1] << " ), ";
            const int cur_hr_trans = hr_trans[i][j];
	    const int toi = trans_order_idx.at(cur_hr_trans - 1) + 1;
            //cout << "mapping cur_hr_trans " << cur_hr_trans << " to order index " << toi << "\n";
	    hrb.miny = std::min(hrb.miny, toi);
	    hrb.maxy = std::max(hrb.maxy, toi);
	  }
	  //std::cout << "\n";
	  
	  if (i < num_hyperrec)
	  {
            hyperrects.push_back(hrb);
	  //}

	    render_area->minx = std::min(render_area->minx, hrb.minx);
	    render_area->miny = std::min(render_area->miny, hrb.miny);
	    render_area->maxx = std::max(render_area->maxx, hrb.maxx);
	    render_area->maxy = std::max(render_area->maxy, hrb.maxy);

	  }

	  std::cout << "hyperrect " << i << " has bounds " << hrb.minx << ", " << hrb.miny << ", " << hrb.maxx << ", " << hrb.maxy << "\n";
	  std::cout << "render_area has bounds " << render_area->minx << ", " << render_area->miny << ", " << render_area->maxx << ", " << render_area->maxy << "\n";
	}

	/*
        for (int i=render_area->miny; i<render_area->maxy; ++i)
        {
          for (int j=render_area->minx; j<=std::max(render_area->maxx, render_area->minx + 4); ++j)
          {
            cout << trans_db.get(i, j);
          }
          cout << "\n";
        }
	*/

	std::ostringstream out_fname_strm;
	out_fname_strm << solution_type_str(solution_type) << "/" << dataset_name.substr(0, dataset_name.length() - 4) << ".";

	std::ostringstream h_reor_fname_strm;
	h_reor_fname_strm << out_fname_strm.str() << std::setw(2) << std::setfill('0') << num_hyperrec << "_h.zoomed.reordered";
	std::string tl_fname(h_reor_fname_strm.str());

        cout << "writing reordered output to file " << tl_fname << "\n";
        trans_db.write_to_trans_file(tl_fname);
	cout << "writing reordered output to image " << tl_fname << "\n";
	if (!zoom)
	{
          render_area = NULL;
	}

	std::ostringstream reor_fname_strm;
	reor_fname_strm << out_fname_strm.str() << "full.reordered";

	// zoomed images
	write_image3(trans_db, tl_fname, &hyperrects, render_area);

	std::cout << "writing full reordered image to " << reor_fname_strm.str() << "\n";
        write_image3(trans_db, reor_fname_strm.str(), NULL, NULL, render_area);

	delete render_area;

//	std::cout << "final database matrix:\n" << trans_db << "\n";
	
  return 0;
}

