#include <assert.h>
#include <errno.h>
#include <iostream>
#include <iterator>
#include <math.h>
#include <stdlib.h>
#include <sys/time.h>
#include <vector>

int main(int argc, char **argv)
{
  if (argc < 3)
  {
    std::cerr << "Usage: " << argv[0] << " dataset_name sample_pct\n";
    exit(1);
  }

  errno = 0;
  const double sample_pct = strtod(argv[2], NULL);
  if (errno != 0)
  {
    perror("could not parse sample_pct");
    exit(1);
  }

  struct timeval tv;
  if (gettimeofday(&tv, 0) != 0)
  {
    perror("gettimeofday");
    exit(1);
  }
  srand48(tv.tv_sec + tv.tv_usec);

  FILE *ds_f;
  if ((ds_f = fopen(argv[1], "rb")) == NULL)
  {
    perror("fopen");
    exit(1);
  }

  char *line = NULL;
  size_t line_len;
  ssize_t bytes_read;
  uint64_t num_trans = 0;
  while ((bytes_read = getline(&line, &line_len, ds_f)) != -1)
  {
    ++num_trans;
  }

  rewind(ds_f);

  //const uint64_t num_outp_lines = static_cast<uint64_t>(round(sample_pct * (double)num_trans / 100.0));
  const uint64_t num_outp_lines = 250;


  uint64_t remaining_lines = num_trans;
  uint64_t remaining_outp_lines = std::min(num_outp_lines, num_trans);

  bytes_read = getline(&line, &line_len, ds_f);
  assert(bytes_read != -1);
  while (remaining_outp_lines > 0)
  { 
    if ((lrand48() % remaining_lines) < remaining_outp_lines)
    {
      std::cout << line;
      --remaining_outp_lines;
    }
    --remaining_lines;
    if (remaining_outp_lines > 0)
    {
      bytes_read = getline(&line, &line_len, ds_f);
      assert(bytes_read != -1);
    }
  }
  
  if (fclose(ds_f) != 0)
  {
    perror("fclose");
    exit(1);
  }

  return 0;
}
