#include <limits>
#include <stdlib.h>
#include "BitVec/BitVec.hh"
#include "types.hh"
#include "write_image.hh"


int main(int argc, char **argv)
{
  if (argc < 2)
  {
    fprintf(stderr, "Usage: %s filename\n", argv[0]);
    exit(1);
  }

  FILE *f;
  if ((f = fopen(argv[1], "rb")) == NULL)
  {
    perror("fopen");
    exit(2);
  }

  char *line = NULL;
  size_t line_size = 0;
  trans_t max_trans = 0;
  item_t min_item = std::numeric_limits<item_t>::max();
  item_t max_item = std::numeric_limits<item_t>::min();

  item_t item;
  int item_str_len;
  char *line_scan;
  while (getline(&line, &line_size, f) != -1)
  {
    ++max_trans;
    line_scan = line;
    while (sscanf(line_scan, "%d%n", &item, &item_str_len) == 1)
    {
      min_item = std::min(min_item, item);
      max_item = std::max(max_item, item);
      
      line_scan += item_str_len;
    }
  }

  rewind(f);

  BitVec bv(max_trans, min_item, max_item);

  trans_t num_trans = 0;
  while (getline(&line, &line_size, f) != -1)
  {
    ++num_trans;
    line_scan = line;
    while (sscanf(line_scan, "%d%n", &item, &item_str_len) == 1)
    {
      bv.set_one(num_trans, item);

      line_scan += item_str_len;
    }
  }

  if (line != NULL)
  {
    free(line);
  }

  if (fclose(f) != 0)
  {
    perror("fclose");
    exit(3);
  }

//  std::cout << bv;

  HyperRectBounds bounds(1, 1, 7, 44);

  write_image3(bv, std::string(argv[1]), NULL, &bounds);

  return 0;
}
