#include <iostream>
#include <iterator>
#include <limits>
#include <math.h>
#include <sstream>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <vector>
#include "write_image.hh"

void write_image3(const BitVec& bv, const std::string out_filename, const std::vector<HyperRectBounds>* hr_bounds, const HyperRectBounds* render_area, const HyperRectBounds* outline_area)
{
  std::cout << "min_item " << bv.min_item() << "\n";
  std::cout << "max_item " << bv.max_item() << "\n";
  std::cout << "num_trans " << bv.num_trans() << "\n";

  const int item_blocks = (bv.max_item() - bv.min_item() + 1);
  std::cout << "item_blocks " << item_blocks << "\n";
  const int trans_blocks = bv.num_trans();
  std::cout << "trans_blocks " << trans_blocks << "\n";

  HyperRectBounds ra;
  ra.minx = bv.min_item();
  ra.maxx = bv.max_item();
  ra.miny = 1;
  ra.maxy = bv.num_trans();

  if (render_area != NULL)
  {
    ra = *render_area;
    std::cout << "assigning passed render area: " << ra.minx << ", " << ra.miny << ", " << ra.maxx << ", " << ra.maxy << "\n";
  }

  std::cout << "using render area: " << ra.minx << ", " << ra.miny << ", " << ra.maxx << ", " << ra.maxy << "\n";
  
  double cols_per_block = (double)(ra.maxx - ra.minx + 1) / (double)item_blocks;
  double rows_per_block = (double)(ra.maxy - ra.miny + 1) / (double)trans_blocks;

  std::cout << "rows_per_block " << rows_per_block << "\n";
  std::cout << "cols_per_block " << cols_per_block << "\n";

  std::vector<unsigned int> p_tmp(item_blocks, 0);
  std::vector<std::vector<unsigned int> > population(trans_blocks, p_tmp);

  for (int i=ra.miny; i<=ra.maxy; ++i)
  {
    for (int j=ra.minx; j<=ra.maxx; ++j)
    {
      if (bv.get(i, j))
      {
	int row_block = j - 1;
	int col_block = i - 1;
        unsigned int *p_loc = &population.at(col_block).at(row_block);
	++*p_loc;
        //std::cout << "bv bit set at " << i << ", " << j << " increased population of " << col_block << ", " << row_block << " to " << *p_loc << "\n";
      }
    }
  }

  //std::cout << "row bucket: ";
  //std::copy(col_bucket.begin(), row_bucket.end(), std::ostream_iterator<uint64_t>(std::cout, ", "));
  //std::cout << "\n";
  
  double density_max = std::numeric_limits<double>::min();
  double density_min = std::numeric_limits<double>::max();

  std::vector<double> d_tmp(p_tmp.size(), 0);
  std::vector<std::vector<double> > density(population.size(), d_tmp);

  std::string of_base = out_filename.substr(0, out_filename.find('.'));
  if (of_base.rfind('/') != std::string::npos)
  {
    of_base = of_base.substr(of_base.rfind('/') + 1);
  }

  std::ostringstream df_fn_strm;
  df_fn_strm << of_base << ".nb_" << item_blocks << "_" << trans_blocks << ".density";
  std::string df_fname(df_fn_strm.str());
  FILE *df;
  if ((df = fopen(df_fname.c_str(), "rb")) != NULL)
  {
    /*
    int num_ret;
    if ((num_ret = fscanf(df, "%lf %lf", &density_min, &density_max)) != 2)
    {
      perror("fscanf density file");
      fprintf(stderr, "fscanf of density file returned %d\n", num_ret);
      exit(1);
    }

    std::cout << "read density_min " << density_min << ", density_max " << density_max << " from " << df_fname << "\n";
    */

    if (fclose(df) != 0)
    {
      perror("fclose density file (from reading)");
      exit(1);
    }
  }

  for (size_t i=0; i<(size_t)density.size(); ++i)
  {
    for (size_t j=0; j<(size_t)density.at(i).size(); ++j)
    {
      const double density_i_j = (double)population[i][j] / (double)(cols_per_block * rows_per_block);
      density[i][j] = density_i_j;
      density_max = std::max(density_max, density_i_j);
      density_min = std::min(density_min, density_i_j);
    }
  }

  if ((df = fopen(df_fname.c_str(), "wb")) == NULL)
  {
    perror("fopen density file for writing");
    fprintf(stderr, "Error: could not open density file %s\n", df_fname.c_str());
    exit(1);
  }

  if (fprintf(df, "%f %f", density_min, density_max) < 0)
  {
    perror("fprintf density file");
    exit(1);
  }

  if (fclose(df) != 0)
  {
    perror("fclose density file (from writing)");
    exit(1);
  }

  std::cout << "wrote density_min " << density_min << ", density_max " << density_max << " to " << df_fname << "\n";

  // normalized density (0.0 - 1.0)
  std::vector<double> dn_tmp(d_tmp.size(), 0);
  std::vector<std::vector<double> > d_norm(density.size(), dn_tmp);
  
  for (size_t i=0; i<density.size(); ++i)
  {
    for (size_t j=0; j<density.at(i).size(); ++j)
    {
      d_norm[i][j] = (density[i][j] - density_min) / (density_max - density_min);
    }
  }

  const int width = 400, height = 400;

  gdImagePtr im;
  if ((im = gdImageCreate(width, height)) == NULL)
  {
    perror("gdImageCreate");
    exit(9);
  }

  int color_white = gdImageColorAllocate(im, 255, 255, 255);
  int color_black = gdImageColorAllocate(im, 0, 0, 0);
  gdImageFilledRectangle(im, 0, 0, width, height, color_white);

  const int graph_w = 370, graph_h = 370;
  const int graph_xpos = 25;
  const int graph_ypos = (width - graph_h - graph_xpos);

  std::string font("/usr/share/fonts/bitstream-vera/Vera.ttf");
  if (access(font.c_str(), F_OK) != 0)
  {
    font = "/usr/share/fonts/truetype/ttf-bitstream-vera/Vera.ttf";
  }

  std::ostringstream xt;
  xt << (ra.maxx - ra.minx + 1) << " items";
  char *xaxis_title;
  if ((xaxis_title = (char*)malloc(xt.str().length() + 1)) == NULL)
  {
    perror("malloc xaxis_title");
    exit(1);
  }
  strcpy(xaxis_title, xt.str().c_str());

  double font_size = 11.0;
  int brect[8];
  char *font_err;
  if ((font_err = gdImageStringFT(NULL, &brect[0], color_black, (char*)font.c_str(), font_size, 0.0, 0, 0, xaxis_title)) != NULL)
  {
    fprintf(stderr, "Error getting x-axis title bounds: %s\n", font_err);
    exit(1);
  }

  int xaxis_xpos = graph_xpos + (graph_w / 2) - ((brect[3] - brect[0]) / 2);
  //int xaxis_ypos = (graph_ypos + graph_h) + (height - (graph_ypos + graph_h)) / 2  - ((brect[6] - brect[0]) / 2);
  int xaxis_ypos = (graph_ypos + graph_h) + ((height - graph_ypos - graph_h - (brect[7] - brect[1])) / 2);
  if ((font_err = gdImageStringFT(im, &brect[0], color_black, (char*)font.c_str(), font_size, 0.0, xaxis_xpos, xaxis_ypos, xaxis_title)) != NULL)
  {
    fprintf(stderr, "Error printing x-axis title: %s\n", font_err);
    exit(1);
  }

  std::ostringstream yt;
  yt << (ra.maxy - ra.miny + 1) << " transactions";
  char *yaxis_title;
  if ((yaxis_title = (char*)malloc(yt.str().length())) == NULL)
  {
    perror("malloc yaxis_title");
    exit(1);
  }
  strcpy(yaxis_title, yt.str().c_str());

  if ((font_err = gdImageStringFT(NULL, &brect[0], color_black, (char*)font.c_str(), font_size, M_PI / 2.0, width, height, yaxis_title)) != NULL)
  {
    fprintf(stderr, "Error getting y-axis title bounds: %s\n", font_err);
    exit(1);
  }

  int yaxis_ypos = (graph_ypos + graph_h) - ((graph_h - (brect[1] - brect[3])) / 2);
  int yaxis_xpos = (graph_xpos - (brect[0] - brect[6])) / 2 + (brect[0] - brect[6]);
  if ((font_err = gdImageStringFT(im, &brect[0], color_black, (char*)font.c_str(), font_size, M_PI / 2.0, yaxis_xpos, yaxis_ypos, yaxis_title)) != NULL)
  {
    fprintf(stderr, "Error printing y-axis title: %s\n", font_err);
    exit(1);
  }


  double w_sf = (double)graph_w / (double)(ra.maxx - ra.minx + 1);
  double h_sf = (double)graph_h / (double)(ra.maxy - ra.miny + 1);

  for (int i=0; i<(ra.maxy - ra.miny + 1); ++i)
  {
    for (int j=0; j<(ra.maxx - ra.minx + 1); ++j)
    {
      //unsigned int d = population.at(i).at(j);
      //int norm_d = (int)(255.0 - (((double)d / (cols_per_block * rows_per_block)) * 255.0));
      int norm_d = (int)(255.0 - (d_norm[i][j] * 255.0));
      if (norm_d < 255)
      {
        //gdImageSetPixel(im, j, i, gdImageColorResolve(im, norm_d, norm_d, norm_d));
        double x1_d = (double)j * w_sf;
        double y1_d = (double)i * h_sf;
        double x2_d = ((double)j + 1.0) * w_sf;
        double y2_d = ((double)i + 1.0) * h_sf;
        int x1 = (int)round(x1_d);
        int y1 = (int)round(y1_d);
        int x2 = (int)round(x2_d);
        int y2 = (int)round(y2_d);
        //std::cout << "drawing block (" << j << "," << i << ") " << norm_d << " item at graph pos " << x1 << ", " << y1 << ", " << x2 << ", " << y2 << "\n";
        int color = gdImageColorResolve(im, norm_d, norm_d, norm_d);
        if (x1 == x2 && y1 != y2)
        {
          //std::cout << "empty rect at " << x1 << ", " << y1 << ", " << x2 << ", " << y2 << "\n";
          gdImageFilledRectangle(im, graph_xpos + x1, graph_ypos + y1, graph_xpos + x2 + 1, graph_ypos + y2, color);
        } else if (y1 == y2 && x1 != x2) {
          gdImageFilledRectangle(im, graph_xpos + x1, graph_ypos + y1, graph_xpos + x2, graph_ypos + y2 + 1, color);
        } else if (x1 == x2 && y1 == y2) {
          gdImageSetPixel(im, x1, y1, color);
        } else {
          gdImageFilledRectangle(im, graph_xpos + x1, graph_ypos + y1, graph_xpos + x2, graph_ypos + y2, color);
        }
      }
    }
  }

  // neatline
  gdImageRectangle(im, graph_xpos - 1, graph_ypos - 1, graph_xpos + graph_w + 1, graph_ypos + graph_h + 1, color_black);

  if (hr_bounds != NULL)
  {
    const int pal_size = 10;
    int palette[pal_size][3] = { {255, 0, 0}, {0, 255, 0}, {0, 0, 255}, {255, 255, 0}, {255, 0, 255}, {0, 255, 255}, {128, 128, 128}, {0, 128, 255}, {128, 255, 0}, {128, 255, 128 } };
  
    gdImagePtr brushes[pal_size];
    for (int i=0; i<pal_size; ++i)
    {
      brushes[i] = gdImageCreate(3, 3);
      int *pal = palette[i % pal_size];
      gdImageColorAllocate(brushes[i], pal[0], pal[1], pal[2]);
    }

    double px_per_item = (double)graph_w / (double)(ra.maxx - ra.minx + 1);
    double px_per_trans = (double)graph_h / (double)(ra.maxy - ra.miny + 1);

    for (int split_hr=0; split_hr<=1; ++split_hr)
    {
      gdImagePtr im_prehr = gdImageCreate(gdImageSX(im), gdImageSY(im));
      gdImageCopy(im_prehr, im, 0, 0, 0, 0, gdImageSX(im), gdImageSY(im));

      int hr_start = (hr_bounds->size() - 1) / 2;
      int hr_end = 0;
      if (split_hr == 1)
      {
        hr_start = hr_bounds->size() - 1;
	      hr_end = hr_bounds->size() / 2;
      }

      for (int i=hr_start; i >= hr_end; --i)
      {
        const HyperRectBounds& hrb = hr_bounds->at(i);
        int x1 = (int)round(px_per_item * (hrb.minx - ra.minx));
        //std::cout << "x1 is round(" << px_per_item << " * " << "(" << hrb.minx << " - " << ra.minx << ")) == " << x1 << "\n";
        int y1 = (int)round(px_per_trans * (hrb.miny - ra.miny));
        int x2 = (int)round(px_per_item * (hrb.maxx - ra.minx + 1));
        int y2 = (int)round(px_per_trans * (hrb.maxy - ra.miny + 1));
        //(double)graph_h * (double)(ra.maxy - ra.miny + 1) / (double)hrb.maxy);
        std::cout << "drawing hyperrect " << i << " bound " << hrb.minx << ", " << hrb.miny << ", " << hrb.maxx << ", " << hrb.maxy << " around graph pos " << x1 << ", " << y1 << ", " << x2 << ", " << y2 << "\n";
  
        gdImageSetBrush(im_prehr, brushes[i % pal_size]);
  
        gdImageRectangle(im_prehr, graph_xpos + x1, graph_ypos + y1, graph_xpos + x2, graph_ypos + y2, gdBrushed);
  
      }
  
      /*
      const int key_xpos = 244;
      const int key_ypos = 10;
      const int key_w = 146;
      const int key_h = 85;
    
      gdImageFilledRectangle(im_prehr, key_xpos, key_ypos, key_xpos + key_w, key_ypos + key_h, gdImageColorResolve(im_prehr, 255, 255, 255));
      gdImageRectangle(im_prehr, key_xpos, key_ypos, key_xpos + key_w, key_ypos + key_h, gdImageColorResolve(im_prehr, 0, 0, 0));
      gdImageStringFT(im_prehr, &brect[0], color_black, (char*)font.c_str(), 9, 0.0, key_xpos + 60, key_ypos + 2 + 12, (char*)"Key");
      for (int i=0; i<5; ++i)
      {
        std::ostringstream txt;
	int which_hr = i + 1 +  (split_hr == 0 ? 0 : 5);
        txt << "Hyperrect " << which_hr << " bound";
        int ypos = key_ypos + 18 + (12 * (i + 1));
        gdImageSetBrush(im_prehr, brushes[i]);
        gdImageRectangle(im_prehr, key_xpos + 5, ypos - 8, key_xpos + 18, ypos - 2, gdBrushed);
        gdImageStringFT(im_prehr, &brect[0], color_black, (char*)font.c_str(), 9, 0.0, key_xpos + 24, ypos, (char*)txt.str().c_str());
      }
      */
      
      std::ostringstream of;
      of << out_filename << "_";
      of << (hr_end + 1) << "-" << (hr_start + 1);
      /*
      if (split_hr == 0)
      {
        of << "1-10";//"1-5";
      } else if (split_hr == 1) {
        of << "11-20";//"6-10";
      }
      */

      eps_image(im_prehr, of.str());
    }

    for (int i=0; i<pal_size; ++i)
    {
      gdImageDestroy(brushes[i]);
    }
    gdImageDestroy(im);
  } else if (outline_area != NULL) {
    gdImagePtr olb = gdImageCreate(3, 3); 
    gdImageColorAllocate(olb, 0, 255, 255);

    gdImageSetBrush(im, olb);

    gdImageRectangle(im, graph_xpos + (int)round(w_sf * (outline_area->minx)) - 3, graph_ypos + (int)round(h_sf * (outline_area->miny)) - 3, graph_xpos + (int)round(w_sf * (outline_area->maxx)) + 3, graph_ypos + (int)round(h_sf * (outline_area->maxy)) + 3, gdBrushed);

    gdImageDestroy(olb);
  }

  if (hr_bounds == NULL)
  {
    eps_image(im, out_filename);
  }
}


void eps_image(gdImagePtr im, const std::string out_filename)
{
  FILE *pngout;
  if ((pngout = fopen((out_filename + ".png").c_str(), "wb")) == NULL)
  {
    perror("fopen pngout file");
    fprintf(stderr, "Error: could not open png file %s to convert into an eps.\n", (out_filename + ".png").c_str());
    exit(4);
  }
  std::cout << "Writing image file " << (out_filename + ".png") << "\n";
  
  gdImagePng(im, pngout);

  if (fclose(pngout) != 0)
  {
    perror("fclose pngout file");
    exit(5);
  }

  gdImageDestroy(im);

  /*
  std::string convert_cmd = "convert " + out_filename + ".png " + out_filename + ".eps";
  std::cout << convert_cmd << "\n";
  int convert_ret = system(convert_cmd.c_str());
  if (WEXITSTATUS(convert_ret) != 0)
  {
    std::cerr << "Error: convert command: " << convert_cmd << " returned " << WEXITSTATUS(convert_ret) << "\n";
    exit(6);
  }
  */

  /*
  std::ostringstream c_top_bot;
  const int c_top_bot_height = 35;
  c_top_bot << "convert -crop " << width << "x" << c_top_bot_height << "+0+";


  std::ostringstream c_top;
  c_top << c_top_bot.str() << 0 << " " << out_filename << " " << eps_fname << "_top.eps";
  int c_top_ret = system(c_top.str().c_str());
  if (WEXITSTATUS(c_top_ret) != 0)
  {
    std::cerr << "Error: convert command: " << c_top.str() << " returned " << WEXITSTATUS(c_top_ret) << "\n";
    exit(8);
  }

  std::ostringstream c_bot;
  c_bot << c_top_bot.str() << height - c_top_bot_height << " " << out_filename << " " << eps_fname << "_bottom.eps";
  int c_bot_ret = system(c_bot.str().c_str());
  if (WEXITSTATUS(c_bot_ret) != 0)
  {
    std::cerr << "Error: convert command: " << c_bot.str() << " returned " << WEXITSTATUS(c_bot_ret) << "\n";
    exit(8);
  }
  */


  /*
  if (unlink(out_filename.c_str()) != 0)
  {
    perror("unlink");
    exit(7);
  }
  */
}

void write_image(const BitVec& bv, const std::string out_filename)
{
  const int width = bv.max_item() - bv.min_item();
  const int height = bv.num_trans();

  gdImagePtr im;
  if ((im = gdImageCreate(width, height)) == NULL)
  {
    perror("gdImageCreate");
    exit(9);
  }
  int color_white = gdImageColorAllocate(im, 255, 255, 255);
  int color_black = gdImageColorAllocate(im, 0, 0, 0);
  gdImageFilledRectangle(im, 0, 0, width, height, color_white);

  for (uint64_t i=1; i<=bv.num_trans(); ++i)
  {
    for (uint64_t j=bv.min_item(); j<=bv.max_item(); ++j)
    {
      if (bv.get(i, j))
      {
        gdImageSetPixel(im, j - bv.min_item(), i - 1, color_black);
      }
    }
  }

  eps_image(im, out_filename);
}



void write_image2(const BitVec& bv, const std::string out_filename)
{
  uint64_t num_cols = (bv.max_item() - bv.min_item() + 1);
  double rows_per_block = ((double)bv.num_trans() - 1.0) / (double)num_cols;
  std::cout << "rows_per_block " << rows_per_block << "\n";

  std::vector<unsigned char> d_tmp(num_cols, 0);
  std::vector<std::vector<unsigned char> > density(num_cols, d_tmp);
  
  for (uint64_t i=1; i<=bv.num_trans(); ++i)
  {
    for (uint64_t j=bv.min_item(); j<=bv.max_item(); ++j)
    {
      if (bv.get(i, j))
      {
	int add_amt = (int)((double)255 * (1.0 / (double)rows_per_block));
        unsigned char *d_loc = &density.at((i - 1) / num_cols).at(j - bv.min_item());
	//std::cout << "adding " << add_amt << " to " << (int)*d_loc;
	*d_loc = std::min((int)*d_loc + add_amt, 255);
	//std::cout << "; density is now " << (int)*d_loc << "\n";
      }
    }
  }

  //std::cout << "row bucket: ";
  //std::copy(col_bucket.begin(), row_bucket.end(), std::ostream_iterator<uint64_t>(std::cout, ", "));
  //std::cout << "\n";

  int width = num_cols, height = num_cols;

  gdImagePtr im;
  if ((im = gdImageCreate(width, height)) == NULL)
  {
    perror("gdImageCreate");
    exit(9);
  }
  int color_white = gdImageColorAllocate(im, 255, 255, 255);
  gdImageFilledRectangle(im, 0, 0, width, height, color_white);

  for (uint64_t i=1; i<num_cols + 1; ++i)
  {
    for (uint64_t j=bv.min_item(); j<=bv.max_item(); ++j)
    {
      if (bv.get(i, j))
      {
        unsigned char d = density.at(i - 1).at(j - bv.min_item());
        gdImageSetPixel(im, j - bv.min_item(), i - 1, gdImageColorResolve(im, d, d, d));
      }
    }
  }

  eps_image(im, out_filename);
}

